@@ -5,3 +5,5 @@
 #undef HAVE_GET_TEXT
 #undef HAVE_LC_MESSAGES
 #undef HAVE_STPCPY
+#undef HAVE_XKB
+#undef HAVE_XEVIE
